//******************************************************************************************
//* Hyperboloid, by Paul Nylander, bugman123.com, 6/22/05
//* runtime: 18 seconds
//******************************************************************************************

camera{location <5,15,-2> look_at <0,0,0> right -x*image_width/image_height up z rotate <0,90,0>}
light_source{<5,15,-2>,1.5}

//Basic Functions
#macro square(X) X*X #end
#macro mod2(a,b) #local c=mod(a,b); (c>0?c:c+b) #end // always return positive
#macro atan3(Y,X) #if(X=0 & Y=0) #local theta=0; #else #local theta=atan2(Y,X); #end theta #end
#macro hue(h) #local X=mod(6*h,1); #local c=<0,0,0>;
 #switch(mod(floor(6*h),6))
  #case(0) #local c=<1,X,0>; #break
  #case(1) #local c=<1-X,1,0>; #break
  #case(2) #local c=<0,1,X>; #break
  #case(3) #local c=<0,1-X,1>; #break
  #case(4) #local c=<X,0,1>; #break
  #case(5) #local c=<1,0,1-X>; #break
 #end
 c
#end

//Complex Functions
#macro Complex(X,Y) <X,Y> #end
#declare I=Complex(0,1);
#macro Re(Z) Z.x #end
#macro Im(Z) Z.y #end
#macro Conjugate(Z) Complex(Re(Z),-Im(Z)) #end
#macro Abs(Z) vlength(Z) #end
#macro Arg(Z) atan2(Im(Z),Re(Z)) #end
#macro Sqr(Z) Complex(square(Re(Z))-square(Im(Z)),2*Re(Z)*Im(Z)) #end
#macro Pow(Z,n) #local r=Abs(Z); #if(r=0) #local z2=Complex(0,0); #else #local theta=n*Arg(Z); #local z2=pow(r,n)*Complex(cos(theta),sin(theta)); #end z2 #end
#macro Sqrt(Z) Pow(Z,1/2) #end
#macro Mult(z1,z2) Complex(Re(z1)*Re(z2)-Im(z1)*Im(z2),Im(z1)*Re(z2)+Re(z1)*Im(z2)) #end
#macro Div(z1,z2) Mult(z1,Pow(z2,-1)) #end
#macro Exp(Z) exp(Re(Z))*Complex(cos(Im(Z)),sin(Im(Z))) #end

//22 Matrix Functions
#macro Matrix(a,b,c,d) array[2][2]{{a,b},{c,d}} #end
#macro Conjugate2(A) Matrix(Conjugate(A[0][0]),Conjugate(A[0][1]),Conjugate(A[1][0]),Conjugate(A[1][1])) #end
#macro Det(A) Mult(A[0][0],A[1][1])-Mult(A[0][1],A[1][0]) #end
#macro Mult2(Z,A) Matrix(Mult(A[0][0],Z),Mult(A[0][1],Z),Mult(A[1][0],Z),Mult(A[1][1],Z)) #end
#macro Div2(A,Z) Matrix(Div(A[0][0],Z),Div(A[0][1],Z),Div(A[1][0],Z),Div(A[1][1],Z)) #end
#macro Inverse(A) Div2(Matrix(A[1][1],-A[0][1],-A[1][0],A[0][0]),Det(A)) #end
#macro Dot(A1,A2) Matrix(Mult(A1[0][0],A2[0][0])+Mult(A1[0][1],A2[1][0]),Mult(A1[0][0],A2[0][1])+Mult(A1[0][1],A2[1][1]),Mult(A2[0][0],A1[1][0])+Mult(A2[1][0],A1[1][1]),Mult(A2[0][1],A1[1][0])+Mult(A1[1][1],A2[1][1])) #end
#macro Reflect(C,M) Dot(Dot(M,Conjugate2(C)),M) #end
#macro ToMatrix(Z,r) Matrix(Z/r,Complex(r-(square(Re(Z))+square(Im(Z)))/r,0),Complex(1/r,0),-Conjugate(Z)/r) #end
#macro ThetaMatrix(theta1,theta2) #if(theta1>theta2) #local theta2=theta2+2*pi; #end
 ToMatrix(sqrt(2/(1+cos(theta2-theta1)))*Exp(I*(theta1+theta2)/2),tan((theta2-theta1)/2))
#end
#macro zcen(A) Div(A[0][0],A[1][0]) #end
#macro rad(A) abs(1/Re(A[1][0])) #end
#macro nth(i,Alist) Matrix(Alist[i][0][0],Alist[i][0][1],Alist[i][1][0],Alist[i][1][1]) #end
#macro setnth(i,Alist,A) #local Alist[i][0][0]=A[0][0]; #local Alist[i][0][1]=A[0][1]; #local Alist[i][1][0]=A[1][0]; #local Alist[i][1][1]=A[1][1]; Alist #end
#macro join(Alist1,Alist2) #local n1=dimension_size(Alist1,1); #local n2=dimension_size(Alist2,1);
 #local Alist3=array[n1+n2][2][2]; #local i=0;
 #while(i<n1) #local Alist3=setnth(i,Alist3,nth(i,Alist1)); #local i=i+1; #end #local i=0;
 #while(i<n2) #local Alist3=setnth(n1+i,Alist3,nth(i,Alist2)); #local i=i+1; #end Alist3
#end

//Parametric Plotting, function f must be defined before this function can be called
#macro ParametricPlot3D(u1,u2,du, v1,v2,dv) #local imax=int((u2-u1)/du); #local jmax=int((v2-v1)/dv);
 #local Mesh=array[imax+1][jmax+1]; #local i=0;
 #while(i<=imax) #local U=u1+i*du; #local j=0; #while(j<=jmax) #local Mesh[i][j]=f(U,v1+j*dv); #local j=j+1; #end #local i=i+1; #end
 DrawMesh(Mesh)
#end
#macro det3(a,b,c,d,e,f,g,h,i) -c*e*g+b*f*g+c*d*h-a*f*h-b*d*i+a*e*i #end // 33 matrix
#macro quad(p1,n1, p2,n2, p3,n3, p4,n4)
 #if(vlength(p3-p1)<vlength(p4-p2))
  smooth_triangle{p1,n1, p2,n2, p3,n3} smooth_triangle{p1,n1, p3,n3, p4,n4}
 #else
  smooth_triangle{p1,n1, p2,n2, p4,n4} smooth_triangle{p2,n2, p3,n3, p4,n4}
 #end
#end
#macro DrawMesh(Mesh)
 #local imax=dimension_size(Mesh,1); #local jmax=dimension_size(Mesh,2); #local normals=array[imax][jmax];
 #local i=0;
 #while(i<imax) #local j=0; #while(j<jmax) #local normals[i][j]=<0,0,0>; #local j=j+1; #end #local i=i+1; #end
 #local i=0;
 #while(i<imax-1) #local j=0; // adapted from Tim Wenclawiak's Make_Normals function
  #while(j<jmax-1)
   #local p0=(Mesh[i][j]+Mesh[i][j+1]+Mesh[i+1][j]+Mesh[i+1][j+1])/4;
   #local N=vcross(Mesh[i+1][j]-p0,Mesh[i+1][j+1]-p0);
   #local E=vcross(Mesh[i+1][j+1]-p0,Mesh[i][j+1]-p0);
   #local S=vcross(Mesh[i][j+1]-p0,Mesh[i][j]-p0);
   #local W=vcross(Mesh[i][j]-p0,Mesh[i+1][j]-p0);
   #local normals[i][j]=normals[i][j]+S+W;
   #local normals[i+1][j]=normals[i+1][j]+N+W;
   #local normals[i][j+1]=normals[i][j+1]+S+E;
   #local normals[i+1][j+1]=normals[i+1][j+1]+N+E;
   #local j=j+1;
  #end
  #local i=i+1;
 #end
 #local i=0;
 #while(i<imax) #local j=0; #while(j<jmax) #local normals[i][j]=vnormalize(normals[i][j]); #local j=j+1; #end #local i=i+1; #end
 #local i=0;
 #while (i<imax-1) #local j=0;
  #while (j<jmax-1)
   quad(
    Mesh[i  ][j  ],normals[i  ][j  ],
    Mesh[i+1][j  ],normals[i+1][j  ],
    Mesh[i+1][j+1],normals[i+1][j+1],
    Mesh[i  ][j+1],normals[i  ][j+1]
   )
   #local j=j+1;
  #end
  #local i=i+1;
 #end
#end                  

//Calculations
#macro index(g,i) int(3*pow(2,g-1)-3+i) #end
#declare gmax=6; // 8 // 12
#declare Tiles=array[index(gmax+1,0)][2][2];
#declare theta1=pi/2; #declare theta2=0.999*pi; #declare theta3=-0.999*pi/2;
#declare Tiles=setnth(0,Tiles,ThetaMatrix(theta1,theta2));
#declare Tiles=setnth(1,Tiles,ThetaMatrix(theta2,theta3));
#declare Tiles=setnth(2,Tiles,ThetaMatrix(theta3,theta1));
#declare i=0;
#while(i<6)
 #declare Tiles=setnth(index(2,i),Tiles,Reflect(nth((mod(i,2)=0?(i<2?1:0):(i>3?1:2)),Tiles),nth(int(i/2),Tiles)));
 #declare i=i+1;
#end
#declare g=2;
#while(g<gmax) #declare i=0;
 #while(i<index(g,3))
  #declare Tiles=setnth(index(g+1,2*i),Tiles,Reflect(nth(index(g-1,floor(i/2)),Tiles),nth(index(g,i),Tiles)));
  #declare Tiles=setnth(index(g+1,2*i+1),Tiles,Reflect(nth(index(g,2*floor(i/2)+mod(i+1,2)),Tiles),nth(index(g,i),Tiles)));
  #declare i=i+1;
 #end
 #declare g=g+1;
#end

//Draw Chains
#declare white=texture{pigment{rgb 0.75} finish{reflection 0.5}}
#declare silver=texture{pigment{rgb 1} finish{brilliance 4 diffuse 0.2 ambient 0.115 specular 1 reflection 0.5 metallic 1}}
#declare gold=texture{pigment{rgb <1.25,0.775,0.375>} finish{brilliance 4 diffuse 0.2 ambient 0.115 specular 1 reflection <0.65,0.56,0.4> metallic 1}}
#declare white=texture{pigment{rgb 0.875} finish{brilliance 1.75 diffuse 0.5 specular 0.25 reflection 0.5 metallic 1}}
#declare pointiness=1; #declare zh=6; #declare Rh=sqrt(square(zh)-1); #declare Rp=(zh-1)/Rh; // trim Poincare disc at Rp<1
#macro f(r,theta) <2*r*cos(theta),2*r*sin(theta),(1+square(r))>/(1-square(r)) #end
#declare hyperboloid=union{ParametricPlot3D(0,Rp,Rp/10, 0,2*pi,2*pi/36)};
#declare t1=0.75;// transparency
union{object{hyperboloid} object{hyperboloid scale <1,1,-1>}
 pigment{radial
  color_map{[0 rgbt <1,0,0,t1>][1/6 rgbt <1,1,0,t1>][1/3 rgbt <0,1,0,t1>][1/2 rgbt <0,1,1,t1>][2/3 rgbt <0,0,1,t1>][5/6 rgbt <1,0,1,t1>][1 rgbt <1,0,0,t1>]}
  rotate <0,90,0>
 }
 finish{reflection 0.4}
}

#declare rtube=0.1; #declare n=dimension_size(Tiles,1);
#declare ring=torus{Rh,2*rtube texture{white} rotate <90,0,0>};
object{ring translate <0,0,-zh>} object{ring translate <0,0,zh>}
#macro approx(x1,x2) (abs(x2-x1)<0.001) #end
#macro mirror(sign,p) <p.x,p.y,sign*p.z> #end
#macro DrawChain(i) #local C=nth(i,Tiles); #local z0=zcen(C); #local r=rad(C);
 #local d=square(r)+square(Rp)-square(Re(z0)); // discriminant
 #if(true) // d>0

  #local R=sqrt(1+square(r));
  #local theta1=acos(Re(z0)/R)-atan(r); #local theta2=theta1+2*atan(r); // angles in main disk... must convert to find chain angles
  #if(!approx(Abs(Exp(I*theta2)-z0),r) | theta1=theta2)
   #local theta2=-acos(Re(z0)/R)+atan(r); #local theta1=theta2-2*atan(r);
  #end

  #local theta1=Arg(Exp(I*theta1)-z0); #local theta2=Arg(Exp(I*theta2)-z0); // convert to find chain angles
  #if(theta1>theta2) #local theta2=theta2+2*pi; #end // could use a swap here instead
  #if(abs(theta2-theta1)>pi) #local t1=theta1; #local theta1=theta2; #local theta2=t1+2*pi; #end // swap and add
  #local theta1=theta1+0.001; #local theta2=theta2-0.001; // prevents singularities
  #local theta=theta1;
  #while(theta<=theta2) #local zp=z0+r*Exp(I*theta); #local rp=Abs(zp);
   #local p1=<2*Re(zp),2*Im(zp),(1+square(rp))>/(1-square(rp)); #local sign=-1;
   #while(sign<=1)
    #if(p1.z<zh) sphere{mirror(sign,p1),rtube texture{white}} #end
    #if(theta>theta1)
     #if(max(p1.z,p2.z)<zh & vlength(p2-p1)>0) cylinder{mirror(sign,p1),mirror(sign,p2),rtube texture{white}} #end
     #if(p1.z<zh & p2.z>zh) cylinder{mirror(sign,p1),mirror(sign,p1+(p2-p1)*(zh-p1.z)/(p2.z-p1.z)),rtube texture{white}} #end
     #if(p1.z>zh & p2.z<zh) cylinder{mirror(sign,p2+(p1-p2)*(zh-p2.z)/(p1.z-p2.z)),mirror(sign,p2),rtube texture{white}} #end
    #end
    #local sign=sign+2;
   #end
   #declare p2=p1; #declare theta=theta+(theta2-theta1)/36;
  #end
 #end
#end
union{
 #declare i=0; #while(i<n) DrawChain(i) #declare i=i+1; #end
 rotate <0,0,180>
}
